Require Import Map.
Require Import TM.

Ltac InvertAll :=
  repeat
    match goal with
      | H: _ /\ _ |- _ => inversion_clear H
      | H: ex _   |- _ => inversion_clear H
    end.

Definition MemPred := Mem -> Prop.

Definition emp_m (M : Mem) : Prop := 
  forall (l : Address), M l = None.

Definition star_m (m m' : MemPred) : MemPred :=
  fun M => exists M1, exists M2, dis_M M1 M2 /\ M = uni_M M1 M2 /\  m M1 /\ m' M2.

Definition imp_m (m m' : MemPred) : Prop :=
  forall M, m M -> m' M.

Lemma star_m_imp : forall m1 m2 m1' m2', imp_m m1 m1' -> imp_m m2 m2' -> imp_m (star_m m1 m2) (star_m m1' m2').
Proof.
intros m1 m2 m1' m2' Himp1 Himp2 M Hstar.
inversion_clear Hstar as [M1 (M2,(Hdis,(Huni,(Hm1,Hm2))))].
exists M1.
exists M2.
auto.
Qed.

Definition prec_m (m : MemPred) : Prop :=
  forall M M' M'', subseteq M' M -> subseteq M'' M -> m M' -> m M'' -> M' = M''.

Definition Assertion := XState -> Prop.

Definition and_a (a a' : Assertion) : Assertion := fun X => a X /\ a' X.

Definition star_am (a : Assertion) (m : MemPred) : Assertion :=
  fun X => match X with
           | (M,(R,k),L) => star_m (fun M1 => a (M1,(R,k),L)) m M
	   end.

Definition star_av (a v : Assertion) : Assertion :=
  fun X => match X with
           | (M,(R,k),L) => star_m (fun M1 => a (M1,(R,k),L)) (fun M2 => v (M2,(R,k), filter_L L k)) M
           end.

Definition imp_a (a a' : Assertion) : Prop := forall X, a X -> a' X.

Definition prec_a (a : Assertion) : Prop :=
  forall M M' M'' R k L, 
    subseteq M' M -> subseteq M'' M -> 
      a (M', (R, k), L) -> a (M'', (R, k), L) -> M' = M''.

Definition freed_a (l : Lock) : Assertion :=
  fun X => match X with
           | (M,(R,k),L) => not_in_dom l L
           end.

Definition owner_a (l : Lock) : Assertion :=
  fun X => match X with
           | (M,(R,k),L) => maps_to L l k
           end.

Definition indom_a (rs : Reg) (w : Word) : Assertion :=
  fun X => match X with
           | (M,(R,k),L) => has_M M (R rs + w)
           end.

Definition ASSUME := XState -> XState -> Prop.

Definition GRT := XState -> XState -> Prop.

Definition liftG (a : Assertion) (G : GRT) (a' : Assertion) : GRT :=
  fun (X X' : XState) =>
    match X with
    | (M, (R, k), L) =>
      match X' with
	(M', (R', k'), L') =>
	  exists M_1, exists M_2, exists M_1', exists M_2',
	    disjUnion M_1 M_2 M /\ disjUnion M_1' M_2' M' /\
	    a (M_1, (R, k), L) /\ a' (M_1', (R', k'), L') /\
	    G (M_1, (R, k), L) (M_1', (R', k'), L')
      end
    end.

Definition presvA (a : Assertion) (A : ASSUME) : Prop :=
  forall (X X' : XState), a X -> A X X' -> a X'.

Definition NI (A A' : ASSUME) (G G' : GRT) : Prop :=
  (forall M M' Ri Ri' i j L L' Rj,
     i <> j -> G (M, (Ri, i), L) (M', (Ri', i), L') -> A' (M, (Rj, j), L) (M', (Rj, j), L')) /\
  (forall M M' Ri Ri' i j L L' Rj,
     i <> j -> G' (M, (Ri, i), L) (M', (Ri', i), L') -> A (M, (Rj, j), L) (M', (Rj, j), L')).
